using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace maAccuMeterNet
{
    class CUsbDriver
    {

        const int DIGCF_PRESENT = 0x2;
        const int DIGCF_ALLCLASSES = 0x4;
        const int DIGCF_DEVICEINTERFACE = 0x10;
        const int BUFFER_SIZE = 256;

        [Flags]
        enum EFileAccess : uint
        {
            GenericRead = 0x80000000,
            GenericWrite = 0x40000000,
            GenericExecute = 0x20000000,
            GenericAll = 0x10000000
        }

        [Flags]
        enum EFileShare : uint
        {
            None = 0x00000000,
            /// <summary>
            /// Enables subsequent open operations on an object to request read access.
            /// Otherwise, other processes cannot open the object if they request read access.
            /// If this flag is not specified, but the object has been opened for read access, the function fails.
            /// </summary>
            Read = 0x00000001,
            /// <summary>
            /// Enables subsequent open operations on an object to request write access.
            /// Otherwise, other processes cannot open the object if they request write access.
            /// If this flag is not specified, but the object has been opened for write access, the function fails.
            /// </summary>
            Write = 0x00000002,
            /// <summary>
            /// Enables subsequent open operations on an object to request delete access.
            /// Otherwise, other processes cannot open the object if they request delete access.
            /// If this flag is not specified, but the object has been opened for delete access, the function fails.
            /// </summary>
            Delete = 0x00000004
        }

        enum ECreationDisposition : uint
        {
            /// <summary>
            /// Creates a new file. The function fails if a specified file exists.
            /// </summary>
            New = 1,
            /// <summary>
            /// Creates a new file, always.
            /// If a file exists, the function overwrites the file, clears the existing attributes, combines the specified file attributes,
            /// and flags with FILE_ATTRIBUTE_ARCHIVE, but does not set the security descriptor that the SECURITY_ATTRIBUTES structure specifies.
            /// </summary>
            CreateAlways = 2,
            /// <summary>
            /// Opens a file. The function fails if the file does not exist.
            /// </summary>
            OpenExisting = 3,
            /// <summary>
            /// Opens a file, always.
            /// If a file does not exist, the function creates a file as if dwCreationDisposition is CREATE_NEW.
            /// </summary>
            OpenAlways = 4,
            /// <summary>
            /// Opens a file and truncates it so that its size is 0 (zero) bytes. The function fails if the file does not exist.
            /// The calling process must open the file with the GENERIC_WRITE access right.
            /// </summary>
            TruncateExisting = 5
        }

        [Flags]
        enum EFileAttributes : uint
        {
            Readonly = 0x00000001,
            Hidden = 0x00000002,
            System = 0x00000004,
            Directory = 0x00000010,
            Archive = 0x00000020,
            Device = 0x00000040,
            Normal = 0x00000080,
            Temporary = 0x00000100,
            SparseFile = 0x00000200,
            ReparsePoint = 0x00000400,
            Compressed = 0x00000800,
            Offline = 0x00001000,
            NotContentIndexed = 0x00002000,
            Encrypted = 0x00004000,
            Write_Through = 0x80000000,
            Overlapped = 0x40000000,
            NoBuffering = 0x20000000,
            RandomAccess = 0x10000000,
            SequentialScan = 0x08000000,
            DeleteOnClose = 0x04000000,
            BackupSemantics = 0x02000000,
            PosixSemantics = 0x01000000,
            OpenReparsePoint = 0x00200000,
            OpenNoRecall = 0x00100000,
            FirstPipeInstance = 0x00080000
        }

        [StructLayout(LayoutKind.Sequential)]
        struct SP_DEVICE_INTERFACE_DATA
        {
            public Int32 cbSize;
            public Guid interfaceClassGuid;
            public Int32 flags;
            private UIntPtr reserved;
        }

        [StructLayout(LayoutKind.Sequential)]
        struct SP_DEVINFO_DATA
        {
            public uint cbSize;
            public Guid ClassGuid;
            public uint DevInst;
            public IntPtr Reserved;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        struct SP_DEVICE_INTERFACE_DETAIL_DATA
        {
            public UInt32 cbSize;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = BUFFER_SIZE)]
            public string DevicePath;
        }

        [StructLayout(LayoutKind.Sequential)]
        struct HIDD_ATTRIBUTES
        {
            public Int32 Size;
            public Int16 VendorID;
            public Int16 ProductID;
            public Int16 VersionNumber;

        }

        [DllImport("hid.dll")]
        static extern void HidD_GetHidGuid(
            out Guid pGuid
            );

        [DllImport("hid.dll", SetLastError = true)]
        static extern Boolean HidD_GetAttributes(
            IntPtr HidDeviceObject,
            ref HIDD_ATTRIBUTES Attributes
            );

        [DllImport("hid.dll", SetLastError = true)]
        static extern Boolean HidD_GetManufacturerString(
            IntPtr HidDeviceObject, 
            ref Byte lpBuffer, 
            Int32 BufferLength
            );

        [DllImport("hid.dll", SetLastError = true)]
        static extern Boolean HidD_GetProductString(
            IntPtr HidDeviceObject,
            ref Byte lpBuffer,
            Int32 BufferLength
            );

        [DllImport("hid.dll")]
        static extern bool HidD_SetFeature(
            IntPtr HidDeviceObject,
            ref Byte lpBuffer,
            Int32 BufferLength
            );

        [DllImport("hid.dll")]
        static extern bool HidD_GetFeature(
            IntPtr HidDeviceObject,
            ref Byte lpBuffer,
            Int32 BufferLength
            );


        [DllImport("setupapi.dll", CharSet = CharSet.Auto)]
        static extern IntPtr SetupDiGetClassDevs(
            ref Guid ClassGuid,
            IntPtr Enumerator,
            IntPtr hwndParent,
            int Flags
            );

        [DllImport(@"setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern bool SetupDiDestroyDeviceInfoList(IntPtr hDevInfo);

        [DllImport("setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern Boolean SetupDiEnumDeviceInterfaces(
           IntPtr hDevInfo,
           IntPtr devInfo,
           ref Guid interfaceClassGuid,
           UInt32 memberIndex,
           ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData
        );

        [DllImport("setupapi.dll", CharSet = CharSet.Auto, SetLastError = true)]
        static extern Boolean SetupDiGetDeviceInterfaceDetail(
           IntPtr hDevInfo,
           ref SP_DEVICE_INTERFACE_DATA deviceInterfaceData,
           ref SP_DEVICE_INTERFACE_DETAIL_DATA deviceInterfaceDetailData,
           UInt32 deviceInterfaceDetailDataSize,
           out UInt32 requiredSize,
           ref SP_DEVINFO_DATA deviceInfoData
        );

        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        static extern IntPtr CreateFile(
           string lpFileName,
           EFileAccess dwDesiredAccess,
           EFileShare dwShareMode,
           IntPtr lpSecurityAttributes,
           ECreationDisposition dwCreationDisposition,
           EFileAttributes dwFlagsAndAttributes,
           IntPtr hTemplateFile);


        static public Boolean UsbHidSetReport(IntPtr handle, Byte[] buffer, int nLen)
        {
            return HidD_SetFeature(handle, ref buffer[0], nLen);
        }

        static public Boolean UsbHidGetReport(IntPtr handle, ref Byte[] buffer, int nLen)
        {
            return HidD_GetFeature(handle, ref buffer[0], nLen);
        }

        static public IntPtr UsbHidOpenDevice(uint unVendorId, string strVendorName,
            uint unProdId, string strProdName)
        {
            Guid guid;
            HidD_GetHidGuid(out guid);
            IntPtr hDeviceInfoList = SetupDiGetClassDevs(ref guid, IntPtr.Zero, IntPtr.Zero, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

            SP_DEVICE_INTERFACE_DATA dia = new SP_DEVICE_INTERFACE_DATA();
            dia.cbSize = Marshal.SizeOf(dia);

            for (int i = 0; ; i++)
            {
                if (SetupDiEnumDeviceInterfaces(hDeviceInfoList, IntPtr.Zero, ref guid, (uint)i, ref dia) != true)
                    break;

                // build a DevInfo Data structure
                SP_DEVINFO_DATA da = new SP_DEVINFO_DATA();
                da.cbSize = (uint)Marshal.SizeOf(da);

                // build a Device Interface Detail Data structure
                SP_DEVICE_INTERFACE_DETAIL_DATA didd = new SP_DEVICE_INTERFACE_DETAIL_DATA();
                if (IntPtr.Size == 8) // for 64 bit operating systems
                    didd.cbSize = 8;
                else
                    didd.cbSize = (uint)(4 + Marshal.SystemDefaultCharSize); // for 32 bit systems

                uint unSize = BUFFER_SIZE;
                SetupDiGetDeviceInterfaceDetail(hDeviceInfoList, ref dia, ref didd, unSize, out unSize, ref da);

                /* attempt opening for R/W -- we don't care about devices which can't be accessed */
                IntPtr handle = CreateFile(didd.DevicePath, EFileAccess.GenericRead | EFileAccess.GenericWrite,
                    EFileShare.Read | EFileShare.Write, IntPtr.Zero,
                    ECreationDisposition.OpenExisting, 0, IntPtr.Zero);

                //invalid handle
                if (handle.ToInt32() == -1)
                    continue;

                HIDD_ATTRIBUTES deviceAttributes = new HIDD_ATTRIBUTES();
                deviceAttributes.Size = Marshal.SizeOf(deviceAttributes);
                HidD_GetAttributes(handle, ref deviceAttributes);

                if (deviceAttributes.VendorID != unVendorId || deviceAttributes.ProductID != unProdId)
                    continue;   /* ignore this device */

                if (strVendorName != "")
                {
                    Byte[] buffer = new Byte[256];
                    if (HidD_GetManufacturerString(handle, ref buffer[0], buffer.Length) != true)
                        continue;

                    string strVendorTmp = System.Text.Encoding.Unicode.GetString(buffer);
                    strVendorTmp = strVendorTmp.TrimEnd( '\0' );
                    if (strVendorName != strVendorTmp)
                        continue;
                }

                if (strProdName != "")
                {
                    Byte[] buffer = new Byte[256];
                    if (HidD_GetProductString(handle, ref buffer[0], buffer.Length) != true)
                        continue;

                    string strProdTmp = System.Text.Encoding.Unicode.GetString(buffer);
                    strProdTmp = strProdTmp.TrimEnd('\0');
                    if (strProdName != strProdTmp)
                        continue;
                }

                /* we have found the device we are looking for! */
                SetupDiDestroyDeviceInfoList(hDeviceInfoList);                
                return handle;
            }

            SetupDiDestroyDeviceInfoList(hDeviceInfoList);

            return IntPtr.Zero;
        }
    }
}
