﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace maAccuMeterNet
{
    public partial class frmMain : Form
    {
        enum eState
        {
            eNotConnected,
            eConnected,
        };

        enum eReport0Type
        {
            eReport0TypeNone
            , eReport0TypeVref
            , eReport0TypeMinVoltage
            , eReport0TypeOcr1A
            , eReport0TypeOcr1B
            , eReport0TypeVolt1a
            , eReport0TypeVolt1b
            , eReport0TypeCurrent1
            , eReport0TypeVolt2a
            , eReport0TypeVolt2b
            , eReport0TypeCurrent2
        };

        IntPtr m_pUsbHandle = IntPtr.Zero;
        eState m_eState = eState.eNotConnected;
        private Mutex m_mtxUsb = new Mutex(); //for USB synchronization

        List<CVoltsItem> m_lstItems = new List<CVoltsItem>();
        bool m_bWorking = false;
        double m_fCapacity1 = 0;
        double m_fCapacity2 = 0;

        public frmMain()
        {
            InitializeComponent();

            OnUpdateGui();

            backgroundWorker1.RunWorkerAsync();
        }

        private Boolean Work()
        {
            return m_bWorking;
        }

        private void UpdateTitle()
        {
            if (m_eState != eState.eNotConnected)
            {
               
                //notifyIcon1.Text = GetTemp().ToString("F3") + "°C";
                Text = "maAccuMeter [Connected]";
            }
            else
            {
                Text = "maAccuMeter [Not connected]";
                //notifyIcon1.Text = "[Not connected]";
            }

            if (m_lstItems.Count > 0)
            {
                TimeSpan timeSpan = m_lstItems[m_lstItems.Count - 1].Date - m_lstItems[0].Date;
                Text += " " + timeSpan.ToString(@"hh\:mm\:ss");
            }
        }

        delegate int RefreshDataIntervalDelegate();
        private int RefreshDataInterval()
        {
            RefreshDataIntervalDelegate del = new RefreshDataIntervalDelegate(OnRefreshDataInterval);
            return (int)Invoke(del);
        }

        private int OnRefreshDataInterval()
        {
            return Decimal.ToInt16( m_numRefreshData.Value);
        }

        delegate void UpdateGuiDelegate();
        private void UpdateGui()
        {
            UpdateGuiDelegate del = new UpdateGuiDelegate(OnUpdateGui);
            Invoke(del);
        }

        private void OnUpdateGui()
        {
            UpdateTitle();

            if (m_lstItems.Count==0)
                return;

            m_lbV1a.Text = m_lstItems[m_lstItems.Count-1].V1a.ToString();
            m_lbV1b.Text = m_lstItems[m_lstItems.Count - 1].V1b.ToString();
            m_lbI1.Text = m_lstItems[m_lstItems.Count - 1].I1.ToString();
            m_lbCapacity1.Text = m_fCapacity1.ToString("F3");

            m_lbV2a.Text = m_lstItems[m_lstItems.Count - 1].V2a.ToString();
            m_lbV2b.Text = m_lstItems[m_lstItems.Count - 1].V2b.ToString();
            m_lbI2.Text = m_lstItems[m_lstItems.Count - 1].I2.ToString();
            m_lbCapacity2.Text = m_fCapacity2.ToString("F3");

            //UpdateIcon();

            AddItemToChart(m_lstItems[m_lstItems.Count-1]);
        }

        private void AddItem(CVoltsItem item)
        {
            //calculate capacity
            if (m_lstItems.Count > 0)
            {
                TimeSpan timeSpan = item.Date - m_lstItems[m_lstItems.Count - 1].Date;

                //avrage_I * second_count / 3600 [mAh]     
                m_fCapacity1 += (double)timeSpan.TotalSeconds / 3600.0 *
                    ((m_lstItems[m_lstItems.Count - 1].I1 + item.I1) / 2.0);

                m_fCapacity2 += (double)timeSpan.TotalSeconds / 3600.0 *
                    ((m_lstItems[m_lstItems.Count - 1].I2 + item.I2) / 2.0);
            }

            m_lstItems.Add(item);
        }

        private void AddItemToChart(CVoltsItem item)
        {
            TimeSpan timeOffsetFirst = m_lstItems[0].Date - DateTime.MinValue;

            if (m_chart.Series[0].Points.Count == 0)
            {
                AddItemToChart(m_lstItems[0].Date - timeOffsetFirst, m_lstItems[0].V1a,
                    m_lstItems[0].V1b, m_lstItems[0].I1, m_lstItems[0].V2a,
                    m_lstItems[0].V2b, m_lstItems[0].I2);             

                //Tag contains last item
                m_chart.Tag = m_lstItems[0];
            }
            else
            {
                //Tag contains last item
                CVoltsItem itemLast = m_chart.Tag as CVoltsItem;
                TimeSpan timeOffset = item.Date - itemLast.Date;
                if (timeOffset.TotalSeconds >= Decimal.ToDouble(m_numRefreshChart.Value) )
                {
                    UInt32 unSumSecond = 0;
                    UInt32 un32AvrV1a = 0;
                    UInt32 un32AvrV1b = 0;
                    UInt32 un32AvrV2a = 0;
                    UInt32 un32AvrV2b = 0;
                    UInt32 un32AvrI1 = 0;
                    UInt32 un32AvrI2 = 0;

                    for (int nIndex = m_lstItems.IndexOf(itemLast) + 1;
                        nIndex < m_lstItems.Count; ++nIndex)
                    {
                        CVoltsItem itemNext = m_lstItems[nIndex];
                        TimeSpan timeSpan = itemNext.Date - itemLast.Date;
                        un32AvrV1a += itemNext.V1a * (UInt32)timeSpan.Seconds;
                        un32AvrV1b += itemNext.V1b * (UInt32)timeSpan.Seconds;
                        un32AvrV2a += itemNext.V2a * (UInt32)timeSpan.Seconds;
                        un32AvrV2b += itemNext.V2b * (UInt32)timeSpan.Seconds;
                        un32AvrI1 += itemNext.I1 * (UInt32)timeSpan.Seconds;
                        un32AvrI2 += itemNext.I2 * (UInt32)timeSpan.Seconds;
                        unSumSecond += (UInt32)timeSpan.Seconds;

                        if (itemNext == item)
                            break;
                    }

                    AddItemToChart(item.Date - timeOffsetFirst, (UInt16)(un32AvrV1a / unSumSecond),
                        (UInt16)(un32AvrV1b / unSumSecond), (UInt16)(un32AvrI1 / unSumSecond),
                        (UInt16)(un32AvrV2a / unSumSecond), (UInt16)(un32AvrV2a / unSumSecond),
                        (UInt16)(un32AvrI2 / unSumSecond));
                
                    m_chart.Tag = item;
                }
            }
        }

        private void AddItemToChart(DateTime time, UInt16 unV1a, UInt16 unV1b, UInt16 unI1, 
            UInt16 unV2a, UInt16 unV2b, UInt16 unI2)
        {
            if (m_chkChartV1a.Checked)
                m_chart.Series["V1a"].Points.AddXY(time, unV1a);
            if (m_chkChartV1b.Checked)
                m_chart.Series["V1b"].Points.AddXY(time, unV1b);
            if (m_chkChartI1.Checked)
                m_chart.Series["I1"].Points.AddXY(time, unI1);

            if (m_chkChartV2a.Checked)
                m_chart.Series["V2a"].Points.AddXY(time, unV2a);
            if (m_chkChartV2b.Checked)
                m_chart.Series["V2b"].Points.AddXY(time, unV2b);
            if (m_chkChartI2.Checked)
                m_chart.Series["I2"].Points.AddXY(time, unI2);
        }

        private Boolean SetReport0Value(IntPtr pUsbHandle, eReport0Type eType, UInt16 unValue)
        {
            bool bResult = false;

            if (pUsbHandle != IntPtr.Zero)
            {
                Byte[] buffer = new Byte[5];
                buffer[0] = 0;
                buffer[1] = (byte)eType;
                buffer[2] = 1;
                buffer[3] = (byte)(unValue / 256);
                buffer[4] = (byte)(unValue % 256);

                m_mtxUsb.WaitOne();
                bResult= CUsbDriver.UsbHidSetReport(pUsbHandle, buffer, 5);
                m_mtxUsb.ReleaseMutex();
            }

            return bResult;
        }

        private Boolean _GetReport0Value(IntPtr pUsbHandle, ref UInt16 unValue)
        {
            if (pUsbHandle != IntPtr.Zero)
            {
                Byte[] buffer = new Byte[5];
                buffer[0] = 0;
                if (CUsbDriver.UsbHidGetReport(pUsbHandle, ref buffer, 5))
                {
                    unValue = (UInt16)(buffer[2] + buffer[1] * 256);
                    return true;
                }
            }

            return false;
        }

        private Boolean _SetReport0Value(IntPtr pUsbHandle, eReport0Type eType)
        {
            if (pUsbHandle != IntPtr.Zero)
            {
                Byte[] buffer = new Byte[5];
                buffer[0] = 0;
                buffer[1] = (byte)eType;
                buffer[2] = 0;
                return CUsbDriver.UsbHidSetReport(pUsbHandle, buffer, 5);
            }

            return false;
        }

        private Boolean GetReport0Value(IntPtr pUsbHandle, eReport0Type eType, ref UInt16 unValue)
        {
            bool bResult = false;

            if (pUsbHandle != IntPtr.Zero)
            {
                m_mtxUsb.WaitOne();
                if (_SetReport0Value(pUsbHandle, eType))
                    bResult = _GetReport0Value(pUsbHandle, ref unValue);
                m_mtxUsb.ReleaseMutex();
            }

            return bResult;
        }

        private Boolean ReadVolts(IntPtr pUsbHandle)
        {
            UInt16 unVolt1A = 0;
            UInt16 unVolt1B = 0;
            UInt16 unVolt2A = 0;
            UInt16 unVolt2B = 0;

            if (GetReport0Value(pUsbHandle, eReport0Type.eReport0TypeVolt1a, ref unVolt1A)
                && GetReport0Value(pUsbHandle, eReport0Type.eReport0TypeVolt1b, ref unVolt1B)
                
                && GetReport0Value(pUsbHandle, eReport0Type.eReport0TypeVolt2a, ref unVolt2A)
                && GetReport0Value(pUsbHandle, eReport0Type.eReport0TypeVolt2b, ref unVolt2B))
            {
                AddItem(new CVoltsItem(unVolt1A, unVolt1B, unVolt2A, unVolt2B));
          
                return true;
            }

            return false;
        }

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            for (; ; )
            {
                switch (m_eState)
                {
                    case eState.eNotConnected:
                        m_pUsbHandle = CUsbDriver.UsbHidOpenDevice(0x16c0, "skalski.biz", 0x05df, "maAccuMeter");
                        if (m_pUsbHandle != IntPtr.Zero)
                        {
                            //set parameters 
                            SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent1,
                                (UInt16)(m_chkBattery1.Checked ? Decimal.ToUInt16(m_numI1.Value) : 0));
                            SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent2,
                                (UInt16)(m_chkBattery2.Checked ? Decimal.ToUInt16(m_numI2.Value) : 0));

                            m_eState = eState.eConnected;
                            UpdateGui();
                        }
                        else
                        {
                            UpdateGui();
                            Thread.Sleep(5000); //try connect after 5sec
                        }

                        break;

                    case eState.eConnected:
                        if ( Work() )
                        {
                            if (m_lstItems.Count > 0)
                            {
                                TimeSpan timeSpan = DateTime.Now - m_lstItems[m_lstItems.Count - 1].Date;
                                if (timeSpan.Seconds < RefreshDataInterval())
                                {
                                    Thread.Sleep(100);
                                    break;
                                }
                            }

                            if (ReadVolts(m_pUsbHandle))
                            {
                                UpdateGui();
                            }
                            else
                            {
                                m_eState = eState.eNotConnected;
                                UpdateGui();
                            }
                        }

                        Thread.Sleep(100);
                        break;
                 
                };
            }
        }

        private void btStart_Click(object sender, EventArgs e)
        {
            m_bWorking = !m_bWorking;

            m_btStart.Text = m_bWorking ? "Stop" : "Start";
        }

        private void btSet_Click(object sender, EventArgs e)
        {
            bool b = SetReport0Value(m_pUsbHandle,
                (eReport0Type)m_cmbReport0Types.SelectedIndex,
                Decimal.ToUInt16(m_numValue.Value));
        }

        private void btGet_Click(object sender, EventArgs e)
        {
            UInt16 unTmp = 0;
            bool b = GetReport0Value(m_pUsbHandle,
                (eReport0Type)m_cmbReport0Types.SelectedIndex,
                ref unTmp);

            if (b)
                m_numValue.Value = unTmp;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            foreach (string s in Enum.GetNames(typeof(eReport0Type)))
                m_cmbReport0Types.Items.Add(s);

            MinimumSize = Size;
        }

        private void ResetChart()
        {
            for (int nIndex = 0; nIndex < m_chart.Series.Count; nIndex++)
                m_chart.Series[nIndex].Points.Clear();

            m_chart.Tag = null;
        }

        private void m_chkChart_CheckedChanged(object sender, EventArgs e)
        {
            ResetChart();

            foreach (CVoltsItem item in m_lstItems)
                AddItemToChart(item);
        }

        private void btReset_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Are you sure to clear all data?", "Reset data", MessageBoxButtons.YesNo)
                != DialogResult.Yes)
                return;

            ResetChart();

            m_lstItems.Clear();

            m_fCapacity1 = 0;
            m_fCapacity2 = 0;
        }

        private void m_numRefreshChart_ValueChanged(object sender, EventArgs e)
        {
            ResetChart();

            foreach (CVoltsItem item in m_lstItems)
                AddItemToChart(item);
        }

        private void m_chkBattery1_CheckedChanged(object sender, EventArgs e)
        {
            SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent1,
                (UInt16)(m_chkBattery1.Checked ? Decimal.ToUInt16(m_numI1.Value) : 0));
        }

        private void m_chkBattery2_CheckedChanged(object sender, EventArgs e)
        {
            SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent2,
                (UInt16)(m_chkBattery2.Checked ? Decimal.ToUInt16(m_numI2.Value) : 0));      
        }

        private void m_numI1_ValueChanged(object sender, EventArgs e)
        {
            if (m_chkBattery1.Checked)
                SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent1,
                    Decimal.ToUInt16(m_numI1.Value));
        }

        private void m_numI2_ValueChanged(object sender, EventArgs e)
        {
            if (m_chkBattery2.Checked)
                SetReport0Value(m_pUsbHandle, eReport0Type.eReport0TypeCurrent2,
                    Decimal.ToUInt16(m_numI2.Value));
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            backgroundWorker1.Dispose();
        }

        private void m_btAbout_Click(object sender, EventArgs e)
        {
            frmAbout dlg = new frmAbout();
            dlg.ShowDialog();
        }
    }
}
